#ifndef ACTIVEOBJECT_HPP
#define ACTIVEOBJECT_HPP


#include <memory>
#include <functional>
#include <list>
#include <pthread.h>
#include <semaphore.h>


//mutex
class Mutex {
public:
    //constructor
    Mutex();

    //destructor
    ~Mutex();

    //lock
    void lock();

    //unlock
    void unlock();

private:
    pthread_mutex_t m_mutex;
    friend class WaitCondition;
};


//wait condition
class WaitCondition {
public:
    //constructor
    WaitCondition();

    //destructor
    ~WaitCondition();

    //signal the condition
    void signal();

    //wait
    void wait(Mutex &mutex);

private:
    pthread_cond_t m_cond;
};


//semaphore
class Semaphore {
public:
    //constructor
    Semaphore();

    //destructor
    ~Semaphore();

    //increment
    void operator ++();

    //decrement
    void operator --();

private:
    sem_t m_sem;
};


//synchronized queue
template <class T> class Queue {
public:
    //add item
    void put(const T &item) {
        m_mutex.lock();
        m_list.push_back(item);
        m_mutex.unlock();
        ++m_semaphore;
    }

    //get item
    void get(T &item) {
        --m_semaphore;
        m_mutex.lock();
        item = m_list.front();
        m_list.pop_front();
        m_mutex.unlock();
    }

private:
    Mutex m_mutex;
    Semaphore m_semaphore;
    std::list<T> m_list;
};


//active object call result.
template <class T = void> class Result {
public:
    //constructor
    Result(const T &v = T()) : m_result(new _Result(v)) {
    }

    //get value
    T get() const {
        return m_result->get();
    }

    //set value
    void set(const T &t) {
        m_result->set(t);
    }

    //get value operator
    operator T () const {
        return get();
    }

    //set value operator
    void operator = (const T &t) {
        set(t);
    }

private:
    class _Result {
    public:
        //constructor
        _Result(const T &v = T()) : m_ok(false), m_value(v) {
        }

        //get value
        T get() {
            m_mutex.lock();
            if (!m_ok) {
                m_cond.wait(m_mutex);
                m_ok = true;
            }
            T result = m_value;
            m_mutex.unlock();
            return result;
        }

        //set value
        void set(const T &v) {
            m_mutex.lock();
            m_value = v;
            m_ok = true;
            m_mutex.unlock();
            m_cond.signal();
        }

    private:
        Mutex m_mutex;
        WaitCondition m_cond;
        bool m_ok;
        T m_value;
    };

    std::shared_ptr<_Result> m_result;
};


//specialization for void
template <> class Result<void> {
public:
    //constructor
    Result() : m_result(new _Result()) {
    }

    //wait for operation
    void get() const {
        return m_result->get();
    }

    //signal operation
    void set() {
        m_result->set();
    }

private:
    class _Result {
    public:
        //constructor
        _Result() : m_ok(false) {
        }

        //get value
        void get() {
            m_mutex.lock();
            if (!m_ok) {
                m_cond.wait(m_mutex);
                m_ok = true;
            }
            m_mutex.unlock();
        }

        //set value
        void set() {
            m_mutex.lock();
            m_ok = true;
            m_mutex.unlock();
            m_cond.signal();
        }

    private:
        Mutex m_mutex;
        WaitCondition m_cond;
        bool m_ok;
    };

    std::shared_ptr<_Result> m_result;
};


//active object
class ActiveObject {
public:
    //constructor
    ActiveObject();

    //destructor
    ~ActiveObject();

protected:
    //put message in queue
    void put(const std::function<void()> &f);

private:
    pthread_t m_thread;
    bool m_loop;
    Queue<std::function<void()>> m_queue;

    //run
    void _run();

    //stop
    void _stop();

    static void *_threadProc(void *p);
};


#endif //ACTIVEOBJECT_HPP
