#include <Windows.h>
#include <cstdio>
#include <ctime>
#include <string>
#include <vector>
#include <cassert>
#include "ActiveObject.hpp"
using namespace std;


//random
int rnd(int i) {
    return rand() % i;
}


//console class; serializes message printing.
class Console : public ActiveObject {
public:
    //printf
    template <class ...A> void printf(A ...a) {
        put(bind(&::printf, a...));
    }
} console;


//dinining philosophers
class DininingPhilosophers {
public:
    //dining table
    class Table : public ActiveObject {
    public:
        //constructor
        Table(int forks) : m_forks(forks) {
        }

        //grab forks
        void grabForks(int left, int right) {
            Result<> leftReq, rightReq;
            put(bind(&Table::_grabForks, this, left, right, leftReq, rightReq));
            leftReq.get();
            rightReq.get();
        }

        //release forks
        void releaseForks(int left, int right) {
            put(bind(&Table::_releaseForks, this, left, right));
        }

    private:
        //fork
        class _Fork {
        public:
            _Fork() : m_used(false) {}

            void addRequest(const Result<> &req) {
                m_requests.push_back(req);
            }

            void beginRequest() {
                if (m_used || m_requests.empty()) return;
                m_used = true;
                m_requests.front().set();
                m_requests.pop_front();
            }

            void endRequest() {
                assert(m_used);
                m_used = false;
            }

        private:
            list<Result<> > m_requests;
            bool m_used;
        };

        //forks
        vector<_Fork> m_forks;

        //grab forks
        void _grabForks(int left, int right, Result<> leftReq, Result<> rightReq) {
            m_forks[left].addRequest(leftReq);
            m_forks[right].addRequest(rightReq);
            m_forks[left].beginRequest();
            m_forks[right].beginRequest();
        }

        //release forks
        void _releaseForks(int left, int right) {
            m_forks[left].endRequest();
            m_forks[right].endRequest();
            m_forks[left].beginRequest();
            m_forks[right].beginRequest();
        }
    };

    //philosopher
    class Philosopher : public ActiveObject {
    public:
        //constructor
        Philosopher(const char *name, Table &table, int leftFork, int rightFork) :
            m_name(name),
            m_table(table),
            m_leftFork(leftFork),
            m_rightFork(rightFork),
            m_loop(true),
            m_timeThinking(0),
            m_timeWaiting(0),
            m_timeEating(0)
        {
        }

        //start
        void start() {
            put(bind(&Philosopher::_start, this));
        }

        //end
        void end() {
            m_loop = false;
            Result<> req;
            put(bind(&Philosopher::_end, this, req));
            req.get();
        }

        //print statistics
        void printStatistics() {
            put(bind(&Philosopher::_printStatistics, this));
        }

    private:
        //name
        const char *m_name;

        //table
        Table &m_table;

        //forks
        int m_leftFork, m_rightFork;

        //loop flag
        bool m_loop;

        //statistics
        clock_t m_timeThinking;
        clock_t m_timeWaiting;
        clock_t m_timeEating;

        //execute loop
        void _start() {
            clock_t start, end;

            while(m_loop) {
                //think
                start = clock();
                console.printf("%s thinking...\n", m_name);
                Sleep(rnd(1000));
                end = clock();
                m_timeThinking += end - start;

                //wait for forks
                start = clock();
                console.printf("%s waiting...\n", m_name);
                m_table.grabForks(m_leftFork, m_rightFork);
                end = clock();
                m_timeWaiting += end - start;

                //eat
                start = clock();
                console.printf("%s eating...\n", m_name);
                Sleep(rnd(1000));
                end = clock();
                m_timeEating += end - start;

                //release the forks
                m_table.releaseForks(m_leftFork, m_rightFork);
            }
        }

        //end
        void _end(Result<> req) {
            req.set();
        }

        //print statistics
        void _printStatistics() {
            console.printf(
                "%s : thinking = %8li, waiting = %8li, eating = %8li\n",
                m_name,
                (m_timeThinking * 1000 / CLOCKS_PER_SEC),
                (m_timeWaiting * 1000 / CLOCKS_PER_SEC),
                (m_timeEating * 1000 / CLOCKS_PER_SEC));
        }
    };

    //main
    static int main() {
        srand(time(0));

        //dining table
        Table table(5);

        //philosophers
        Philosopher aristotle("Aristotle", table, 0, 1);
        Philosopher kant     ("Kant     ", table, 1, 2);
        Philosopher spinoza  ("Spinoza  ", table, 2, 3);
        Philosopher marx     ("Marx     ", table, 3, 4);
        Philosopher russell  ("Russell  ", table, 4, 0);

        //start
        aristotle.start();
        kant.start();
        spinoza.start();
        marx.start();
        russell.start();

        //wait for key
        getchar();

        //end
        aristotle.end();
        kant.end();
        spinoza.end();
        marx.end();
        russell.end();

        //print stats
        aristotle.printStatistics();
        kant.printStatistics();
        spinoza.printStatistics();
        marx.printStatistics();
        russell.printStatistics();

        return 0;
    }
};


int main() {
    DininingPhilosophers::main();
    return 0;
}
